Attribute VB_Name = "CurveFunctions"
Option Explicit

Private Const INFINITY As Double = 1E+38
Private Const BIG_VALUE As Double = 100000000#
Private Const TINY As Double = 0.0000001

Private Const G_A_MIN As Double = -10
Private Const G_A_MAX As Double = 10
Private Const G_B_MIN As Double = -10
Private Const G_B_MAX As Double = 10
Private Const G_C_MIN As Double = -10
Private Const G_C_MAX As Double = 10

' Return the function for x, a, b, and c.
Public Function F(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
    On Error GoTo FError
    F = a + b * Exp(c * x)
    Exit Function

FError:
    F = INFINITY
End Function

' dF/da
Private Function dFda(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
    dFda = 1
End Function

' dF/db
Private Function dFdb(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
    On Error GoTo dFdbError
    dFdb = Exp(c * x)
    Exit Function

dFdbError:
    dFdb = INFINITY
End Function

' dF/dc
Private Function dFdc(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
    On Error GoTo dFdcError
    dFdc = b * Exp(c * x) * x
    Exit Function

dFdcError:
    dFdc = INFINITY
End Function

' Return the error squared.
Public Function ErrorSquared(ByVal PtX As Collection, ByVal PtY As Collection, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
Dim total As Double
Dim pt As Integer
Dim dy As Double

    On Error GoTo ErrorSquaredError
    total = 0
    For pt = 1 To PtX.Count
        dy = PtY(pt) - F(PtX(pt), a, b, c)
        total = total + dy * dy
    Next pt
    ErrorSquared = total
    Exit Function

ErrorSquaredError:
    ErrorSquared = INFINITY
End Function

' dE/da
Private Function dEda(ByVal PtX As Collection, ByVal PtY As Collection, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
Dim total As Double
Dim i As Integer

    On Error GoTo dEdaError
    total = 0
    For i = 1 To PtX.Count
        total = total + _
            2 * (F(PtX(i), a, b, c) - PtY(i)) * _
            dFda(PtX(i), a, b, c)
    Next i
    dEda = total
    Exit Function

dEdaError:
    dEda = INFINITY
End Function

' dE/db
Private Function dEdb(ByVal PtX As Collection, ByVal PtY As Collection, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
Dim total As Double
Dim i As Integer

    On Error GoTo dEdbError
    total = 0
    For i = 1 To PtX.Count
        total = total + _
            2 * (F(PtX(i), a, b, c) - PtY(i)) * _
            dFdb(PtX(i), a, b, c)
    Next i
    dEdb = total
    Exit Function

dEdbError:
    dEdb = INFINITY
End Function

' dE/dc
Private Function dEdc(ByVal PtX As Collection, ByVal PtY As Collection, ByVal a As Double, ByVal b As Double, ByVal c As Double) As Double
Dim total As Double
Dim i As Integer

    On Error GoTo dEdcError
    total = 0
    For i = 1 To PtX.Count
        total = total + _
            2 * (F(PtX(i), a, b, c) - PtY(i)) * _
            dFdc(PtX(i), a, b, c)
    Next i
    dEdc = total
    Exit Function

dEdcError:
    dEdc = INFINITY
End Function

' Find a good curve fit.
Public Function FindGoodFit(ByVal PtX As Collection, ByVal PtY As Collection, _
    ByRef best_a As Double, ByRef best_b As Double, ByRef best_c As Double, _
    ByVal num_steps As Integer, ByVal max_iterations As Integer) As Double
Dim test_a As Double
Dim test_b As Double
Dim test_c As Double
Dim g_a_step As Double
Dim g_b_step As Double
Dim g_c_step As Double
Dim best_error2 As Double
Dim test_error2 As Double
Dim a As Double
Dim b As Double
Dim c As Double

    best_a = 0
    best_b = 0
    best_c = 0

    g_a_step = (G_A_MAX - G_A_MIN) / (num_steps - 1)
    g_b_step = (G_B_MAX - G_B_MIN) / (num_steps - 1)
    g_c_step = (G_C_MAX - G_C_MIN) / (num_steps - 1)

    best_error2 = INFINITY
    For a = G_A_MIN To G_A_MAX Step g_a_step
        For b = G_B_MIN To G_B_MAX Step g_b_step
            For c = G_C_MIN To G_C_MAX Step g_c_step
                test_error2 = FollowGradient(PtX, PtY, a, b, c, _
                    test_a, test_b, test_c, max_iterations)
                If (test_error2 < best_error2) Then
                    best_a = test_a
                    best_b = test_b
                    best_c = test_c
                    best_error2 = test_error2
                End If
            Next c
        Next b
    Next a

    FindGoodFit = Sqr(best_error2)
End Function

' Starting at (X, Y), follow the error
' function's gradient to try to improve it.
Private Function FollowGradient(ByVal PtX As Collection, ByVal PtY As Collection, _
    ByVal a As Double, ByVal b As Double, ByVal c As Double, _
    ByRef best_a As Double, ByRef best_b As Double, ByRef best_c As Double, _
    ByVal max_iterations As Integer) As Double
Const CUTOFF_ERROR As Double = 1#
Dim dist As Double
Dim best_error2 As Double
Dim iteration As Integer
Dim na As Double
Dim nb As Double
Dim nc As Double
Dim length As Double
Dim va As Double
Dim vb As Double
Dim vc As Double
Dim test_error2 As Double

    ' Set the initial distance to move.
    dist = 0.25

    ' Start with the initial point.
    best_a = a
    best_b = b
    best_c = c
    best_error2 = ErrorSquared(PtX, PtY, a, b, c)

    For iteration = 1 To max_iterations
        ' Get the gradient at this point.
        na = dEda(PtX, PtY, a, b, c)
        nb = dEdb(PtX, PtY, a, b, c)
        nc = dEdc(PtX, PtY, a, b, c)

        ' Normalize it.
        On Error Resume Next
        length = na * na + nb * nb + nc * nc
        If err.Number <> 0 Then
            ' The values are too big. Return the result so far.
            FollowGradient = best_error2
            Exit Function
        End If
        On Error GoTo 0
        length = Sqr(length)

        ' If the length is too small, return the result so far.
        If (length < TINY) Then
            FollowGradient = best_error2
            Exit Function
        End If
        na = -na / length
        nb = -nb / length
        nc = -nc / length

        ' Try moving along the gradient.
        va = na * dist
        vb = nb * dist
        vc = nc * dist
        test_error2 = ErrorSquared(PtX, PtY, a + va, b + vb, c + vc)

        ' Try increasingly smaller vectors
        ' until we find an improvement.
        Do While (test_error2 > best_error2)
            dist = dist / 2
            If (dist < TINY) Then
                ' We're not moving far enough. Stop.
                ' Return the result so far.
                FollowGradient = best_error2
                Exit Function
            End If
            va = na * dist
            vb = nb * dist
            vc = nc * dist
            test_error2 = ErrorSquared(PtX, PtY, a + va, b + vb, c + vc)
        Loop

        ' Hopefully at this point we have found an improvement.
        a = a + va
        b = b + vb
        c = c + vc
        best_a = a
        best_b = b
        best_c = c
        best_error2 = test_error2

        ' If the error is small enough, return.
        If (best_error2 < CUTOFF_ERROR) Then
            FollowGradient = best_error2
            Exit Function
        End If
    Next iteration

    ' Return the error.
    FollowGradient = best_error2
End Function
